/*--------------------------------------------------------------------------*/
// Sample Application													
//																			
// Chapter 4																
//																			
// Learn DirectX in 24 Hours												
// by Robert Dunlop															
//																			
// Copyright (C) 1999														
/*--------------------------------------------------------------------------*/

//------ Include Files ------//

#include "stdafx.h"
#define INITGUID
#include <stdio.h>
#include <ddraw.h>
#include <mmsystem.h>

//------ Window Class Information ------//

static char szClass[] = "XmplHr4Class";
static char szCaption[] = "Example - Hour 4";

//------ Define Position Limits ------//

#define MIN_POS				0.0
#define MAX_POS				50000.0

//------ Global Interface Pointers ------//

LPDIRECTDRAW7 lpDD=NULL;
LPDIRECTDRAWSURFACE7 lpDDSPrimary=NULL;
LPDIRECTDRAWSURFACE7 lpDDSBack=NULL;

//------ DirectDraw Surfaces for Object Storage ------//

LPDIRECTDRAWSURFACE7 back_surf[3]={NULL,NULL,NULL};

//------ Define Starting Position and Speed ------//

double x_pos=25000.0;		// player position
double move_rate=0.0;		// player motion

//------ Error Return String ------//

const char *ErrStr=NULL;	

//------ Error Messages ------//

const char Err_Reg_Class[]			= "Error Registering Window Class";
const char Err_Create_Win[]			= "Error Creating Window";
const char Err_DirectDrawCreate[]	= "DirectDrawCreate FAILED";
const char Err_Query[]				= "QueryInterface FAILED";
const char Err_Coop[]				= "SetCooperativeLevel FAILED";
const char Err_CreateSurf[]			= "CreateSurface FAILED";
const char Err_GetBack[]			= "Error Retrieving Back Buffer";
const char Err_LoadBMP[]			= "Error Loading Image";
const char Err_DispMode[]			= "Error Setting Display Mode";
const char Err_LoadImage[]			= "Error Loading Image";

//------ Function to Load a Bitmap into a DirectDraw Surface ------//

LPDIRECTDRAWSURFACE7 bitmap_surface(LPCTSTR file_name,RECT *dims=NULL)
{
	HDC hdc;
	HBITMAP bit;
	LPDIRECTDRAWSURFACE7 surf;

	// load the interface bitmap

	bit=(HBITMAP) LoadImage(NULL,file_name,IMAGE_BITMAP,0,0,
								LR_DEFAULTSIZE|LR_LOADFROMFILE);
	if (!bit) 

		// failed to load, return failure to caller

		return NULL;

	// get bitmap dimensions

	BITMAP bitmap;
    GetObject( bit, sizeof(BITMAP), &bitmap );
	int surf_width=bitmap.bmWidth;
	int surf_height=bitmap.bmHeight;

	// create surface

	HRESULT ddrval;
	DDSURFACEDESC2 ddsd;
	ZeroMemory(&ddsd,sizeof(ddsd));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT ;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN; 
	ddsd.dwWidth = surf_width;
	ddsd.dwHeight = surf_height; 

	// attempt to create surface

	ddrval=lpDD->CreateSurface(&ddsd,&surf,NULL);

	// created ok?

	if (ddrval!=DD_OK) {

		// no, release the bitmap and return failure to caller

		DeleteObject(bit);
		return NULL;

	} else {

		// yes, get a DC for the surface

		surf->GetDC(&hdc);

		// generate a compatible DC

		HDC bit_dc=CreateCompatibleDC(hdc);

		// blit the interface to the surface

		SelectObject(bit_dc,bit);
		BitBlt(hdc,0,0,surf_width,surf_height,bit_dc,0,0,SRCCOPY);

		// release the DCs

		surf->ReleaseDC(hdc);
		DeleteDC(bit_dc);

		// save the dimensions if rectangle pointer provided

		if (dims) {
			dims->left=0;
			dims->top=0;
			dims->right=surf_width;
			dims->bottom=surf_height;
		}
	}

	// clear bitmap 

	DeleteObject(bit);

	// return pointer to caller

	return surf;
}

//------ Cleanup Function to Release Objects ------//

#define SafeRelease(x) if (x) { x->Release(); x=NULL;}

void Cleanup()
{
	// release loaded image surfaces

	SafeRelease(back_surf[0]);
	SafeRelease(back_surf[1]);
	SafeRelease(back_surf[2]);

	// release DirectDraw interfaces

	SafeRelease(lpDDSPrimary);
	SafeRelease(lpDD);

	// display error if one thrown

	if (ErrStr) {
		MessageBox(NULL, ErrStr, szCaption, MB_OK);
		ErrStr=NULL;
	}
}

//------ Windows Message Handler ------//

LRESULT CALLBACK 
WindowProc(HWND hWnd, unsigned uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {

        case WM_DESTROY:
			Cleanup();
            PostQuitMessage(0);
            break;

        case WM_KEYDOWN: 
            switch (wParam) 
            { 
                case VK_LEFT: 
                    
                    // Process the LEFT ARROW key. 
                    
					if (move_rate>-600.0)
						move_rate-=120.0;
                    break; 
 
                case VK_RIGHT: 
                    
                    // Process the RIGHT ARROW key. 
                     
					if (move_rate<600.0)
						move_rate+=120.0;
                    break; 
 
				case VK_ESCAPE:

					// exit the program on escape

					DestroyWindow(hWnd);
					break;

                // Process other non-character keystrokes. 
                 
                default: 
                    break; 
            }

        default:
            return DefWindowProc(hWnd, uMsg, wParam, lParam);
    }
	return 0L;
}

//------ Function to load object images and set up layers ------//

BOOL load_images()
{
	// load the background images

	back_surf[0]=bitmap_surface("city1.bmp");
	if (!back_surf[0]) {
		ErrStr=Err_LoadImage;
		return FALSE;
	}
	back_surf[1]=bitmap_surface("city2.bmp");
	if (!back_surf[1]) {
		ErrStr=Err_LoadImage;
		return FALSE;
	}
	back_surf[2]=bitmap_surface("city3.bmp");
	if (!back_surf[2]) {
		ErrStr=Err_LoadImage;
		return FALSE;
	}

	// return success to caller

	return TRUE;
}

//------ Function to Initialize DirectDraw and the Application ------//

static BOOL Init(HINSTANCE hInstance, int nCmdShow)
{
    WNDCLASS                    wc;
    HRESULT                     ddrval;

    // Set up and register window class

    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = (WNDPROC) WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = sizeof(DWORD);
    wc.hInstance = hInstance;
    wc.hIcon = NULL;
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH) GetStockObject(BLACK_BRUSH);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = szClass;
    if (!RegisterClass(&wc)) {
		ErrStr=Err_Reg_Class;
        return FALSE;
	}

    // Get dimensions of display

    int ScreenWidth = GetSystemMetrics(SM_CXSCREEN);
    int ScreenHeight = GetSystemMetrics(SM_CYSCREEN);

    // Create a window and display

	HWND hWnd;

    hWnd = CreateWindow(szClass,							// class
                        szCaption,							// caption
						WS_VISIBLE|WS_POPUP,				// style 
						0,									// left
						0,									// top
						ScreenWidth,						// width
						ScreenHeight,						// height
                        NULL,								// parent window
                        NULL,								// menu 
                        hInstance,							// instance
                        NULL);								// parms
    if (!hWnd) {
		ErrStr=Err_Create_Win;
        return FALSE;
	}
    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);

    // Create the main DirectDraw object

    ddrval = DirectDrawCreateEx(NULL, (VOID **) &lpDD, IID_IDirectDraw7, NULL);
    if (ddrval != DD_OK) {
		ErrStr=Err_DirectDrawCreate;
		return FALSE;
	}

    // Set our cooperative level

    ddrval = lpDD->SetCooperativeLevel(hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );
    if (ddrval != DD_OK) {
		ErrStr=Err_Coop;
		return FALSE;
	}

	// Set the display mode

	ddrval = lpDD->SetDisplayMode( 640, 480, 16, 0, 0);
	if (ddrval !=DD_OK) {
		ErrStr=Err_DispMode;
		return FALSE;
	}

    // Create the primary surface with 1 back buffer

    DDSURFACEDESC2 ddsd;
	ZeroMemory(&ddsd,sizeof(ddsd));
    ddsd.dwSize = sizeof( ddsd );
    ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
                          DDSCAPS_FLIP | 
                          DDSCAPS_COMPLEX;
    ddsd.dwBackBufferCount = 1;
    ddrval = lpDD->CreateSurface( &ddsd, &lpDDSPrimary, NULL );
	if (ddrval!=DD_OK) {
		ErrStr=Err_CreateSurf;
		return FALSE;
	}

	// Fetch back buffer interface

	DDSCAPS2 ddscaps;
	ZeroMemory(&ddscaps,sizeof(ddscaps));
	ddscaps.dwCaps=DDSCAPS_BACKBUFFER;
	ddrval=lpDDSPrimary->GetAttachedSurface(&ddscaps,&lpDDSBack);
	if (ddrval!=DD_OK) {
		ErrStr=Err_GetBack;
		return FALSE;
	}

	// load the images an set up the layers

	if (!load_images()) 

		// return with error if failed

		return FALSE;

	// return success to caller

	return TRUE;
}

//------ Function to Draw a Frame ------//

void render_frame()
{
	RECT rct;			// source rectangle of blit
	long tile_pos;		// position of viewable tile

	// loop through the tiles

	for (int i=0;i<3;i++) {
	
		// calculate position of viewable tile

		tile_pos=400*i+((long) (x_pos)/1200)*1200;
		if (tile_pos+1200<x_pos+640) tile_pos+=1200;

		// is this object on the screen?

		if (tile_pos+400>=x_pos) {

			// yes, calculate the offset to the left edge of the screen

			int screen_x=tile_pos-x_pos;

			// clip the object if necessary

			rct.left=0;
			rct.top=0;
			rct.right=400;
			rct.bottom=480;

			// test for clipping

			if (screen_x<0) {
				rct.left-=screen_x;
				screen_x=0;
			} else if (screen_x+rct.right>640) {
				rct.right=640-screen_x;
			}

			// draw the object to the screen

			lpDDSBack->BltFast(screen_x,0,back_surf[i],&rct,DDBLTFAST_WAIT);
		}
	}

	// flip to the primary surface

	lpDDSPrimary->Flip(0,DDFLIP_WAIT); 
}

//------ Application Loop ------//

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	MSG msg;                 // message from queue
	LONGLONG cur_time;       // current time
	BOOL notDone=TRUE;       // flag for thread completion
	DWORD  time_count=16;    // ms per frame, default if no performance counter
	LONGLONG perf_cnt;       // performance timer frequency
	BOOL perf_flag=FALSE;    // flag determining which timer to use
	LONGLONG next_time=0;    // time to render next frame
	LONGLONG last_time=0;	 // time of previous frame
	double time_elapsed;	 // time since previous frame
	double time_scale;		 // scaling factor for time

	// initialize the application, exit on failure

    if (!Init(hInstance, nCmdShow)) {
		Cleanup();
        return FALSE;
	}

	// is there a performance counter available? 

	if (QueryPerformanceFrequency((LARGE_INTEGER *) &perf_cnt)) { 

		// yes, set time_count and timer choice flag 

		perf_flag=TRUE;
		time_count=perf_cnt/60;
		QueryPerformanceCounter((LARGE_INTEGER *) &next_time); 
		time_scale=1.0/perf_cnt;

	} else { 

		// no performance counter, read in using timeGetTime 

		next_time=timeGetTime(); 
		time_scale=0.001;
	} 

	// save time of last frame

	last_time=next_time;

	// run till completed 

	while (notDone) { 

		// is there a message to process? 

		if (PeekMessage( &msg, NULL, 0, 0, PM_REMOVE)) { 

			// yes, is it a quit message? 

			if (msg.message==WM_QUIT) 

				notDone=FALSE; 

			// dispatch the message 

			TranslateMessage(&msg);
			DispatchMessage(&msg); 

		} else {

			// use the appropriate method to get time 
			// and calculate elapsed time since last frame

			if (perf_flag) 
				QueryPerformanceCounter((LARGE_INTEGER *) &cur_time); 
			else 
				cur_time=timeGetTime(); 

			// is it time to render the frame? 

			if (cur_time>next_time) { 

				// calculate elapsed time

				time_elapsed=(cur_time-last_time)*time_scale;

				// save frame time

				last_time=cur_time;
				
				// move the screen position

				x_pos+=move_rate*time_elapsed;
				if (x_pos<MIN_POS) {
					x_pos=MIN_POS;
					move_rate=0;
				}
				if (x_pos>MAX_POS) {
					x_pos=MAX_POS;
					move_rate=0;
				}

				// yes, render the frame 

				render_frame();

				// set time for next frame 

				next_time = cur_time + time_count; 
			}
		}
	}

	// exit returning final message

    return (msg.wParam);
}

